// Device-specific include file.
#include <p33fxxxx.h>

// Standard C library headers.
#include <math.h>
#include <stdlib.h>
#include <string.h>

// Utility functions.
#include <libpic30.h>

#include "vdc.h"
#include "vdc_drawing.h"

// Configuration.

_FOSC(OSCIOFNC_ON & POSCMD_XT)
_FOSCSEL(FNOSC_PRIPLL)
_FWDT(FWDTEN_OFF)

// Constants.
#define M_PI (3.1415926535897932384626433832795f)

// Types.
typedef struct {
	int x;
	int y;
} point2d;

typedef struct {
	int x;
	int y;
	int z;
} point3d;

// Draws a swirly line pattern.
void demo_swirl(void) {

	float angle, sub_angle;
	unsigned int amplitude;
	unsigned int pass;

	vdc_clear();

	for (pass = 0; pass < 2; ++pass) {
		// Draw the pattern.		
		for (amplitude = 0, angle = 0; amplitude < 320; amplitude += 4, angle += 0.08f) {
			for (sub_angle = 0; sub_angle < 2 * M_PI; sub_angle += M_PI * 0.5f) {
				vdc_draw_line(
					160 + amplitude * sin(angle + sub_angle), 120 + amplitude * cos(angle + sub_angle),
					160 + amplitude * sin(angle + sub_angle + M_PI * 0.5f), 120 + amplitude * cos(angle + sub_angle + M_PI * 0.5f)
				);
			}
		}
		if (pass == 0) {
			// Delay for two seconds.
			vdc_delay_frames(vdc_fps * 2);
			// Next loop, we'll clear the screen.
			vdc_set_plot_mode(VDC_PLOT_MODE_CLEAR);
		}
	}

	// Put the plotting mode back the way it was.
	vdc_set_plot_mode(VDC_PLOT_MODE_SET);
}

// Draws a spinning cube.
void demo_cube(void) {
	

	// Source X, Y and Z of each point2d of the cube.
	int x, y, z;

	// Rotation in X and Y.
	float rx = 0, ry = 0;
	
	// Computed sin/cos of the rotated X and Y.
	float sin_rx, cos_rx, sin_ry, cos_ry;
	
	// Transformed X, Y and Z.
	float tx, ty, tz;

	// Stores the eight projected vertices.
	point2d vertices[8];

	// There's always a need for one of these.
	unsigned int i;

	// Size of the cube.
	float rs = 0;
	float s;

	int demo_timer = 315 * 4;

	while (demo_timer-- > 0) {

		// Calculate the sin/cos of the rotated X and Y.
		sin_rx = sin(rx);
		cos_rx = cos(rx);
		sin_ry = sin(ry);
		cos_ry = cos(ry);

		// Calculate the size of the cube.
		s = 140.0f * (1.0f - cos(rs));
	
		// Calculate each vertex in turn.
		i = 0;
		for (x = -1; x <= +1; x += 2) {
			for (y = -1; y <= +1; y += 2) {
				for (z = -1; z <= +1; z+= 2) {
	
					// Calculate the transformed vertex.
					tx = y * cos_rx - x * sin_rx;
					ty = -x * cos_rx * sin_ry - y * sin_rx * sin_ry - z * cos_ry;
	  				tz = 3 - x * cos_rx * cos_ry - y * sin_rx * cos_ry + z * sin_ry;
	
					// Project and store.
					vertices[i].x = 160 + (int)(tx * s / tz);
					vertices[i].y = 120 + (int)(ty * s / tz);
					
					// Advance to the next vertex.
					++i;
				}
			}
		}
		
		vdc_wait_vblank();

		// Clear the LCD.
		vdc_clear();
		
		// One face.
		vdc_draw_line(vertices[0].x, vertices[0].y, vertices[1].x, vertices[1].y);
		vdc_draw_line(vertices[1].x, vertices[1].y, vertices[3].x, vertices[3].y);
		vdc_draw_line(vertices[3].x, vertices[3].y, vertices[2].x, vertices[2].y);
		vdc_draw_line(vertices[2].x, vertices[2].y, vertices[0].x, vertices[0].y);
	
		// Opposite face.
		vdc_draw_line(vertices[4].x, vertices[4].y, vertices[5].x, vertices[5].y);
		vdc_draw_line(vertices[5].x, vertices[5].y, vertices[7].x, vertices[7].y);
		vdc_draw_line(vertices[7].x, vertices[7].y, vertices[6].x, vertices[6].y);
		vdc_draw_line(vertices[6].x, vertices[6].y, vertices[4].x, vertices[4].y);
	
		// Connecting the two faces.
		for (i = 0; i < 4; ++i) {
			vdc_draw_line(vertices[i].x, vertices[i].y, vertices[i + 4].x, vertices[i + 4].y);
		}

		switch (vdc_output_mode) {
			case VDC_OUTPUT_TV:
				rs += 0.02f / 4.0f;
				ry += 0.03f / 4.0f;
				rx += 0.05f / 4.0f;
				break;
			case VDC_OUTPUT_LCD:
				rs += 0.02f;
				ry += 0.03f;
				rx += 0.05f;
				vdc_delay_frames(3);
				demo_timer -= 3;
				break;
		}
	}
	vdc_clear();
}

// Draws a Sierpinksi triangle.
void demo_sierpinski(void) {
	
	float x, y, n;
	unsigned int steps = 15000;

	vdc_clear();
	
	srand(0);

	x = rand() / (float)RAND_MAX;
	y = rand() / (float)RAND_MAX;
	
	while (steps--) {
		n = rand() / (float)RAND_MAX;
		if (n <= 1.0f / 3.0f) {
			x *= 0.5f;
			y *= 0.5f;
		} else if (n <= 2.0f / 3.0f) {
			x = (0.5f + x) * 0.5f;
			y = (1.0f + y) * 0.5f;
		} else {
			x = (1.0f + x) * 0.5f;
			y *= 0.5f;
		}
		vdc_set_pixel(x * 320, (1.0f - y) * 240);
	}

	vdc_delay_frames(vdc_fps);
}

// Draws a Mandelbrot fractal. Based on code by Ken Silverman.
void demo_mandelbrot(void) {

	vdc_draw_line(0,0,319,0);
	vdc_draw_line(0,239,319,239);
	vdc_draw_line(0,0,0,239);
	vdc_draw_line(319,0,319,239);

	// Device (x,y) coordinates.
	int x, y;
	
	// Used to calculate the colour at each point2d.
	float mx, ox, my, oy, xx, yy;

	float hundredth = 1.0f / 100.0f;
	
	// Stores the working colour.
	unsigned int c;

	// Used to dither the results. Taken from http://en.wikipedia.org/wiki/Ordered_dithering
	unsigned int threshold_map[4][4] = {
		{  1,  9,  3, 11 },
		{ 13,  5, 15,  7 },
		{  4, 12,  2, 10 },
		{ 16,  8, 14,  6 }
	};

	for (y = 0; y < 240; ++y) {
		vdc_auto_init();
		for (x = 0; x < 320; ++x) {
	
			// Calculate coordinates of point2d to sample in the fractal.
			ox = mx = (x - 160) * hundredth - 0.7f;
			oy = my = (y - 120) * hundredth;

			// Calculate the colour at the point2d.
			for (c = 0; c < 16; ++c) {
				xx = mx * mx;
				yy = my * my;
				if (xx + yy >= 4) break;
				my = mx * my * 2.0f + oy;
				mx = xx - yy + ox;
			}
			
			// Colour the pixel.
			if (c + threshold_map[x & 3][y & 3] > 16) {
				vdc_set_pixel(x, y);
			}
		}
	}
}

// Flips the screen upside down four times.
void demo_flip(void) {

	unsigned int top_row, bottom_row, column;
	unsigned int* top_ptr;
	unsigned int* bottom_ptr;
	unsigned int flip_count;
	
	for (flip_count = 0; flip_count < 4; ++flip_count) {
		vdc_wait_vblank();
		for (top_row = 0, bottom_row = 239; top_row < 120; ++top_row, --bottom_row) {
			top_ptr = vdc_frame_buffer + 20 * top_row;
			bottom_ptr = vdc_frame_buffer + 20 * bottom_row;
			for (column = 0; column < 20; ++column) {
				unsigned int s = *top_ptr;
				*(top_ptr++) = *bottom_ptr;
				*(bottom_ptr++) = s;
			}
		}
		vdc_delay_frames(vdc_fps / 2);
	}

}

// Draws a circular wipe pattern on the screen.
void demo_circle_wipe(void) {
	unsigned int r, jiggle;
	for (r = 0; r <= 220; ++r) {
		vdc_auto_init();
		for (jiggle = 0; jiggle < 2; ++jiggle) {
			vdc_draw_circle(160 + jiggle, 120, r);
		}
		if (r > 20) {
			vdc_set_plot_mode(VDC_PLOT_MODE_CLEAR);
			for (jiggle = 0; jiggle < 2; ++jiggle) {
				vdc_draw_circle(160 + jiggle, 120, r - 20);
			}
			vdc_set_plot_mode(VDC_PLOT_MODE_SET);
		}
	}
}

// Draw a clock hand at a specified angle and with a specified length.
void demo_clock_draw_hand(float angle, int length) {

	int arrowhead_angle_offset;

	// Main part of the hand.
	vdc_draw_line(160, 120, 160 + length * sin(angle), 120 - length * cos(angle));
	
	// Draw the arrowhead.
	for (arrowhead_angle_offset = -1; arrowhead_angle_offset <= +1; arrowhead_angle_offset += 2) {
		vdc_draw_line(
			160 + length * sin(angle),
			120 - length * cos(angle),
			160 + length * 0.85f * sin(angle + arrowhead_angle_offset * 0.1f),
			120 - length * 0.85f * cos(angle + arrowhead_angle_offset * 0.1f)
		);
	}
}

// Draws the clock demo.
void demo_clock(void) {

	int jitter;
	int hand_position;

	vdc_clear();

	// Draw the outer dial.
	for (jitter = -1; jitter <= 0; ++jitter) {
		// Outer ring.
		vdc_draw_circle(160 + jitter, 120, 118);
		vdc_draw_circle(160, 120 + jitter, 118);
		// Inner ring.
		vdc_draw_circle(160 + jitter, 120, 105);
		vdc_draw_circle(160, 120 + jitter, 105);
	}
	
	// Draw the marks around the dial.
	for (hand_position = 0; hand_position < 60; ++hand_position) {
		float tick_length = (hand_position % 5) ? 97 : 90;
		float angle = (hand_position / 30.0f) * M_PI;
		vdc_draw_line(
			160 + 100 * sin(angle),
			120 - 100 * cos(angle),
			160 + tick_length * sin(angle),
			120 - tick_length * cos(angle)
		);
	}

	for (hand_position = 55; hand_position <= 60; ++hand_position) {
		
		// Draw the current hand position.
		demo_clock_draw_hand((hand_position / 30.0f) * M_PI, 88);
		// Delay one second.
		vdc_delay_frames(vdc_fps);

		// Have we finished?
		if (hand_position == 60) break;

		// Erase the old hand before drawing the next one.
		vdc_set_plot_mode(VDC_PLOT_MODE_CLEAR);
		demo_clock_draw_hand((hand_position / 30.0f) * M_PI, 88);
		vdc_set_plot_mode(VDC_PLOT_MODE_SET);
	}
}

// Slides the image off the bottom of the LCD.
void demo_slide(void) {
	int row;
	unsigned char* vdc_frame_buffer_byte = (unsigned char*)vdc_frame_buffer;
	for (row = 0; row < 240; row += 2) {
		vdc_wait_vblank();
		memmove(vdc_frame_buffer_byte + (row + 2) * 40, vdc_frame_buffer_byte + row * 40, 9600 - (row + 2) * 40);
		memset(vdc_frame_buffer_byte + row * 40, 0, 40 * 2);
	}
}

#define STAR_COUNT 400

void demo_stars(void) {

	point3d stars[STAR_COUNT];

	int star;
	point2d star_transformed;
	point2d previous_star_transformed;
	int speed;

	for (star = 0; star < STAR_COUNT; ++star) stars[star].z = 1;

	vdc_clear();

	int frames_run = 0;
	int stars_drawn = 0;
	int stars_to_draw = 0;

	do {

		// We haven't drawn any stars.
		stars_drawn = 0;

		// Wait for vblank or a particular number of frames.
		// The LCD requires slower, choppier animation due to its slow response.
		switch (vdc_output_mode) {
			case VDC_OUTPUT_TV:
				speed = 5;
				vdc_wait_vblank();
				++frames_run;
				break;
			case VDC_OUTPUT_LCD:
				speed = 20;
				vdc_delay_frames(4);
				frames_run += 4;
				break;
		}
	
		// Increase the number of stars to draw.
		if (stars_to_draw < STAR_COUNT) {
			stars_to_draw = frames_run;
			if (stars_to_draw > STAR_COUNT) stars_to_draw = STAR_COUNT;
		}

		// Move and draw each star.
		for (star = 0; star < stars_to_draw; ++star) {

			previous_star_transformed.x = stars[star].x / stars[star].z + 160;
			previous_star_transformed.y = stars[star].y / stars[star].z + 120;
			
			if ((stars[star].z -= speed) < 50 || previous_star_transformed.x < 0 || previous_star_transformed.x > 319 || previous_star_transformed.y < 0 || previous_star_transformed.y > 239) {
				vdc_clear_pixel(previous_star_transformed.x, previous_star_transformed.y);
				if (frames_run < 1000) {
					// We're still animating, so add a star.
					stars[star].x = ((rand() & 0x1FF) - 256) << 7;
					stars[star].y = ((rand() & 0x1FF) - 256) << 7;
					stars[star].z = 300 + (rand() & 511);
					++stars_drawn;
				} else {
					// Don't bother to redraw the star.
					stars[star].z = 1;
				}
			} else {
				star_transformed.x = stars[star].x / stars[star].z + 160;
				star_transformed.y = stars[star].y / stars[star].z + 120;
				vdc_clear_pixel(previous_star_transformed.x, previous_star_transformed.y);
				vdc_set_pixel(star_transformed.x, star_transformed.y);
				++stars_drawn;
			}
		}
	} while (stars_drawn > 0);
	
	// Ensure the LCD is clear.
	vdc_clear();
}

int main(void) {

	// Configure PLL prescaler, PLL postscaler, PLL divisor
	
	// Fosc=Fin*(M/(N1*N2))
	PLLFBD = 30; // M = 32
	CLKDIVbits.PLLPOST = 0; // N1=2
	CLKDIVbits.PLLPRE = 0; // N2=2

	// Initialise the VDC.
	vdc_auto_init();

	// Run the demos.
	demo_clock();
	demo_slide();
	demo_stars();
	demo_sierpinski();
	demo_flip();
	demo_circle_wipe();
	demo_swirl();
	demo_cube();
	demo_mandelbrot();
	for(;;) vdc_wait_vblank();

}
