// Device-specific include file.
#include <p33fxxxx.h>

// VDC definitions.
#include "vdc.h"

// Stores a 320x240 monochrome bitmap.
unsigned int __attribute__((far)) vdc_frame_buffer[4800];

// Stores a single scanline temporarily (used for DMA transfers).
unsigned int vdc_scanline_buffer[21] __attribute__((space(dma)));

// Points at the current part of the display that's being sent to the display device.
volatile unsigned int* vdc_frame_scan_ptr;

// Stores the current scanline number.
volatile int vdc_active_line_counter;

// Stores the current frame rate of the display refresh.
int vdc_fps;

// Pointer to the current LCD/TV retrace function.
volatile unsigned int vdc_retrace_function_ptr;

// (TV specific) Counter for the number of times to repeat the current TV retrace function.
volatile unsigned int vdc_tv_retrace_function_repeat_count;

// (TV specific) Offsets the display in the X axis.
int vdc_tv_line_offset_x = 82;

// (TV specific) Offsets the display in the Y axis.
int vdc_tv_line_offset_y = 40;

// (TV specific) Initialises the TV retracing routine.
extern void vdc_tv_retrace_init(void);

// (LCD specific) Initialises the LCD retracing routine.
extern void vdc_lcd_retrace_init(void);

// Stores the current output mode.
enum vdc_output_mode_t vdc_output_mode = -1;

// (TV specific) Cleared if the TV is not detected during a retrace.
volatile int vdc_tv_connected;

// De-initialise the TV output code.
void vdc_tv_deinit(void) {

	// Return pin RB5 to an input.
	TRISB |= (1 << 5);

	T1CONbits.TON = 0; // Switch timer 1 off.
	T2CONbits.TON = 0; // Switch timer 2 off.

	OC1CONbits.OCM = 0b000; // Switch output capture unit 1 off.

	// Stop generating timer 1 interrupts.
	IEC0bits.T1IE = 0; // Disable the interrupt.
	IFS0bits.T1IF = 0; // Clear the interrupt flag.
	
	// Switch off the two DMA channels.
	DMA0CONbits.CHEN = 0;
	DMA1CONbits.CHEN = 0;

	_RP4R = 0b00000; // Return pin RP4 to its default use.
	_RP5R = 0b00000; // Return pin RP5 to its default use.
	RPINR20bits.SCK1R = 0b1111; // Pull SCK1 high.

	// Switch off SPI1.
	SPI1STATbits.SPIEN = 0;

	vdc_tv_connected = 0;
}

// De-initialise the LCD output code.
void vdc_lcd_deinit(void) {

	// Switch the LCD off.
	LATB = 0;
	
	T1CONbits.TON = 0; // Switch timer 1 off.

	// Stop generating timer 1 interrupts.
	IEC0bits.T1IE = 0; // Disable the interrupt.
	IFS0bits.T1IF = 0; // Clear the interrupt flag.
	
	// Disconnect the LCD data pins.
	TRISB |= 0b1;
	TRISA |= 0b10011;	
}

// Initialise the TV output code.
void vdc_tv_init(void) {
	
	// Ensure a suitable retrace function is loaded.
	vdc_tv_retrace_init();

	// Reset the scan pointer.
	vdc_frame_scan_ptr = vdc_frame_buffer;
	vdc_active_line_counter = 0;

	// Pins RB4, RB5 and RB6 are outputs.
	TRISB &= ~((1 << 4) | (1 << 5) | (1 << 6));

	// Configure timer 1 to trigger interrupts to constantly retrace the TV.
	T1CONbits.TCKPS = 0b00; // No prescaler.
	PR1 = (FCY / 1000000UL * 64) - 1; // Trigger the interrupt when the counter reaches this value.
	TMR1 = 0; // Reset the timer.
	T1CONbits.TON = 1; // Switch on the timer.

	// Configure timer 2 to generate the clock
	TMR2 = 0;
	PR2 = 4;
	T2CONbits.TON = 1;

	OC1CONbits.OCTSEL = 0;
	OC1CONbits.OCM = 0b101;
	OC1R = 0;
	OC1RS = 2;
	_RP4R = 0b10010; // Output OC1 on pin RP4.
	
	SPI1CON1bits.MSTEN = 0; // Slave transmitter mode.

	SPI1CON1bits.MODE16 = 1; // 16-bit mode.

	SPI1CON1bits.DISSCK = 1; // Disable the SCK pin.
		
	_RP5R = 0b00111; // Output SDO1 on pin RP5.
	RPINR20bits.SCK1R = 4; // Input SCK1 on pin RP4.

	DMACS0 = 0;

	DMA0REQbits.IRQSEL = 0b0001010;
	DMA0PAD = (unsigned int)&SPI1BUF;
	DMA0CONbits.DIR = 1;
	DMA0CONbits.MODE = 0b01;
	DMA0STA = __builtin_dmaoffset(vdc_scanline_buffer);
	DMA0CNT = 20;

	DMA1REQbits.IRQSEL = 0b0001010;
	DMA1PAD = (unsigned int)&SPI1BUF;
	DMA1CONbits.DIR = 0;
	DMA1CONbits.MODE = 0b01;
	DMA1STA = __builtin_dmaoffset(vdc_scanline_buffer);
	DMA1CNT = 20;
	
	// Enable the timer 1 interrupt.
	IFS0bits.T1IF = 0; // Clear the interrupt flag.
	IEC0bits.T1IE = 1; // Enable the interrupt.

	vdc_fps = 50;
	
	vdc_output_mode = VDC_OUTPUT_TV;

	// Assume a TV is connected.
	vdc_tv_connected = !0;
}

void vdc_lcd_init(void) {

	// Ensure a suitable retrace function is loaded.
	vdc_lcd_retrace_init();

	// Reset the scan pointer.
	vdc_frame_scan_ptr = vdc_frame_buffer;
	vdc_active_line_counter = 0;

	// Force the LCD control pins low and make the LCD data pins outputs.
	LATB = 0;
	TRISB &= ~0b1;
	TRISA &= ~0b10011;

	// Configure timer 1 to trigger interrupts to constantly retrace the LCD.
	T1CONbits.TCKPS = 0b00; // No prescaler.
	PR1 = (((unsigned long)FCY / (240)) / 55) - 1; // Trigger the interrupt when the counter reaches this value.
	
	IFS0bits.T1IF = 0; // Clear the interrupt flag.
	IEC0bits.T1IE = 1; // Enable the interrupt.
	
	TMR1 = 0; // Reset the timer.
	T1CONbits.TON = 1; // Switch on the timer.

	// Switch the LCD on.
	LATB |= (1 << 11);

	vdc_fps = 55;

	vdc_output_mode = VDC_OUTPUT_LCD;

}

// Initialises the VDC library.
void vdc_init(enum vdc_output_mode_t vdc_output_mode) {
	
	// Ensure that LCD control pins and TV sync are driven low.
	LATB &= 0b0000011111011111;
	TRISB &= 0b0000011111011111;
	
	// Set interrupt priorities.
	IPC0bits.T1IP = 7; // Timer 1 has highest priority.
	IPC2bits.SPI1EIP = 6; // SPI1 has second-highest priority.
	
	switch (vdc_output_mode) {
		case VDC_OUTPUT_TV:
			vdc_lcd_deinit();
			vdc_tv_init();
			break;
		case VDC_OUTPUT_LCD:
			vdc_tv_deinit();
			vdc_lcd_init();
			break;
	}

}

// Checks whether a TV is connected or not. May only be updated once a frame.
int vdc_tv_is_connected(void) {
	int result = 0;
	switch (vdc_output_mode) {
		default:
			TRISB &= ~(1 << 6);
		case VDC_OUTPUT_LCD:
			// Drive sync high.
			LATB |= (1 << 6);
			Nop();
			Nop();
			Nop();
			// Sample the picture pin; if it's high (no load) then there's no TV, if it's low (loaded) there is a TV.
			result = !(PORTB & (1 << 5));
			LATB &= ~(1 << 6);
			break;
		case VDC_OUTPUT_TV:
			result = vdc_tv_connected;
			break;
	}
	return result;
}

// Automatically (re)initises the VDC to output to the TV (if connected) or the LCD (if not).
void vdc_auto_init(void) {
	if (vdc_tv_is_connected()) {
		if (vdc_output_mode != VDC_OUTPUT_TV) vdc_init(VDC_OUTPUT_TV);
	} else {
		if (vdc_output_mode != VDC_OUTPUT_LCD) vdc_init(VDC_OUTPUT_LCD);
	}
}

// Waits for the display to finish retracing and enter the vertical blank region.
void vdc_wait_vblank(void) {
	vdc_auto_init();
	switch (vdc_output_mode) {
		case VDC_OUTPUT_TV:
			while (vdc_active_line_counter != 240);
			break;
		case VDC_OUTPUT_LCD:
			while (vdc_active_line_counter);
			break;
	}
}

// Delay for a particular number of frames.
void vdc_delay_frames(int frames) {
	while (frames--) {
		vdc_auto_init();
		switch (vdc_output_mode) {
			case VDC_OUTPUT_TV:
				while (vdc_active_line_counter != 240);
				while (vdc_active_line_counter == 240);
				break;
			case VDC_OUTPUT_LCD:
				while (vdc_active_line_counter);
				while (!vdc_active_line_counter);
				break;
		}
	}
}
